unit GLControl;

// Simple control to render OpenGL content.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OpenGL12;

type
  TViewport = record
    Left, Top,
    Width, Height: Integer;
  end;

  TGLControl = class(TWinControl)
  private
    FRenderingContext: HGLRC;
    FDeviceContext: HDC; 
    FViewport: TViewport;
    FOnRender,
    FOnSetup: TNotifyEvent;
    FBackground: TColor;
    FCanRender: Boolean;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); Message WM_ERASEBKGND;
    procedure WMPaint(var Message: TWMPaint); Message WM_PAINT;
    procedure WMSize(var Message: TWMSize); Message WM_SIZE;
    procedure SetBackground(const Value: TColor);
    procedure ApplyBackground;
    procedure ApplyPerspective;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure ClearBuffers;
    procedure DestroyWnd; override;
  public
    constructor Create(AOwner: TComponent); override;

    procedure DisableContext;
    procedure EnableContext;
    procedure RenderScene;

    property RenderingContext: HGLRC read FRenderingContext;
  published
    property Anchors;
    property Align;
    property Background: TColor read FBackground write SetBackground;
    property Constraints;
    property DragCursor;
    property DragMode;
    property Enabled;
    property HelpContext;
    property Hint;

    property PopupMenu;
    property Visible;

    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnRender: TNotifyEvent read FOnRender write FOnRender;
    property OnSetup: TNotifyEvent read FOnSetup write FOnSetup;
  end;

procedure Register;

//----------------------------------------------------------------------------------------------------------------------

implementation

//----------------------------------------------------------------------------------------------------------------------

procedure Register;

begin
  RegisterComponents('OpenGL', [TGLControl]);
end;

//----------------- TGLControl -----------------------------------------------------------------------------------------

procedure TGLControl.ClearBuffers;

begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
end;

//----------------------------------------------------------------------------------------------------------------------

constructor TGLControl.Create(AOwner: TComponent);

begin
  inherited;
  Width := 150;
  Height := 150;
  FCanRender := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.CreateParams(var Params: TCreateParams);

begin
  inherited;
  with Params do
  begin
    Style := Style or WS_CLIPCHILDREN or WS_CLIPSIBLINGS;
    WindowClass.Style :=  CS_VREDRAW or CS_HREDRAW;
  end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.ApplyPerspective;

begin
  with FViewport do
  begin
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity;
    gluPerspective(35, Width / Height, 1, 100);
    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, Width, Height);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.CreateWnd;

var
  Dummy: HPALETTE;
  
begin
  inherited;

  FDeviceContext := GetDC(Handle);
  FRenderingContext := CreateRenderingContext(FDeviceContext, [opDoubleBuffered], 32, 0, 0, 0, 0, Dummy);
  if FRenderingContext = 0 then
    raise Exception.Create('Erzeugung des Renderingkontexts schlug fehl.');

  ActivateRenderingContext(FDeviceContext, FRenderingContext);
  if not GL_VERSION_1_1 then
  begin
    DestroyRenderingContext(FRenderingContext);
    raise Exception.Create('Diese OpenGL version ist veraltet.');
  end;

  // Set the viewport to the entire window size.
  with FViewPort do
  begin
    Left := 0;
    Top := 0;
    Width := Self.Width;
    Height := Self.Height;
    ApplyPerspective;
  end;

  ApplyBackground;

  if Assigned(FOnSetup) then
    FOnSetup(Self);
  DeactivateRenderingContext;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.DestroyWnd;

begin
  ReleaseDC(Handle, FDeviceContext);
  
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.RenderScene;

// This is the actual paint routine. All initialization must already have been done.
// The event FOnRender is triggered to let the application draw whatever is needed.

begin
  ActivateRenderingContext(FDeviceContext, FRenderingContext);
  ClearBuffers;
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  if Assigned(FOnRender) then
    FOnRender(Self);
  SwapBuffers(FDeviceContext);
  DeactivateRenderingContext;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.ApplyBackground;

var
  Color: COLORREF;
  R, G, B: Single;

begin
  ActivateRenderingContext(FDeviceContext, FRenderingContext);
  Color := ColorToRGB(FBackground);
  R := GetRValue(Color) / 255;
  G := GetGValue(Color) / 255;
  B := GetBValue(Color) / 255;
  glClearColor(R, G, B, 1);
  DeactivateRenderingContext;
  Refresh;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.SetBackground(const Value: TColor);

begin
  if FBackground <> Value then
  begin
    FBackground := Value;
    if not (csLoading in ComponentState) then
      ApplyBackground;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.WMEraseBkgnd(var Message: TWMEraseBkgnd);

begin
  Message.Result := 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.WMPaint(var Message: TWMPaint);

var
  PS: TPaintStruct;

begin
  BeginPaint(Handle, PS);
  if FCanRender then
    RenderScene;
  EndPaint(Handle, PS);
  Message.Result := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.WMSize(var Message: TWMSize);

begin
  // Update viewport (here always the entire window).
  with FViewPort do
  begin
    Width := Message.Width;
    Height := Message.Height;
    if Height = 0 then
      Height := 1;
  end;

  if (FRenderingContext <> 0) and  FCanRender then
  begin
    ActivateRenderingContext(FDeviceContext, FRenderingContext);
    ApplyPerspective;
    DeactivateRenderingContext;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.DisableContext;

begin
  FCanRender := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TGLControl.EnableContext;

begin
  if not FCanRender then
  begin
    FCanRender := True;
    if FRenderingContext <> 0 then
    begin
      ActivateRenderingContext(FDeviceContext, FRenderingContext);
      ApplyPerspective;
      DeactivateRenderingContext;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
